'''Defines exceptions used by the step system.
'''
from ..context import Context
from ..logging import Logger

class StepExecutorExceptionBase(RuntimeError):
	'''Base class for all StepExecutor exceptions.
	'''
	pass

class StepExecutorError(StepExecutorExceptionBase):
	'''Raised when an error occurs in the step executor.
	'''
	pass

class PackageError(StepExecutorError):
	'''Runtime errors caused by the package manager.
	'''
	pass

class ShellError(StepExecutorError):
	'''Runtime errors caused by the shell.
	'''
	pass

class CommandExecutionError(ShellError):
	'''A ShellManager.command() call returned a nonzero
	exit code. exit_code will contain the command's exit code.
	'''
	def __init__(self, message, exit_code, *args):
		self.message = message
		self.exit_code = exit_code
		super(CommandExecutionError, self).__init__(message, exit_code, *args)

# Warning exceptions.
# these are unusual situations that
# won't crash the program, but probably will
# break the intended workflow.
# Unless we're developing you probably want to halt
# when these are generated,
# so they throw unless --permit-warnings is set.
class StepExecutorWarning(StepExecutorExceptionBase):
	'''Raised when an warning occurs in the step executor.
	'''
	pass

class PackageWarning(StepExecutorWarning):
	'''Runtime warnings caused by the package manager.
	'''
	pass

class ShellWarning(StepExecutorWarning):
	'''Runtime warnings caused by the shell.
	'''
	pass

class Exceptions(object):
	'''General exception-related
	operations.
	'''

	@classmethod
	def raiseWarning(cls, context, ex):
		'''If `context.permit_warnings` is False,
		raises `ex`.
		Raises RuntimeError if `context.permit_warnings`
		doesn't exist.
		'''
		#Sanity check.
		assert isinstance(context, Context)
		if context.permit_warnings is None:
			raise RuntimeError("context.permit_warnings exists but isn't set. Is context uninitialized?")

		if not context.permit_warnings:
			raise ex

	@classmethod
	def raiseAndLogStepWarning(cls, logger, context, warningString):
		'''Logs a warning string and raises it as
		a StepExecutorWarning if context.permit_warnings is set.
		'''
		assert isinstance(logger, Logger)
		logger.warning(warningString)
		cls.raiseWarning(context, StepExecutorWarning(warningString))



'''
Defines a log renderer that performs its
rendering by HTTP POSTing the log message
to a specified address.
'''

class HttpPoster(object):
	'''HTTP POSTs text logs to a given address.
	'''

	def __init__(self, apiAddress):
		self._apiAddress = apiAddress
		raise NotImplementedError()

	def render(self, logText):
		'''POSTs the log text to the poster's API address.

		:param str logText: The text to be displayed.
		'''
		raise NotImplementedError()



'''Unit tests for log module.
'''

import unittest
from ...setup import Logger, LogLevel

#...How do we test any of this in an automated way?
class LogTest(unittest.TestCase):
	'''Unit tests for log module.
	'''

	def testNormalMethods(self):
		'''Displays all non-contextual methods.
		'''
		testLogger = Logger()

		print("")
		print("You should see the following:")
		print("\t * An entry for info messages")
		print("\t * An entry for warning messages")
		print("\t * An entry for error messages")
		print("\t * An entry for success messages")
		testLogger.info("Info message.")
		testLogger.warning("Warning message.")
		testLogger.error("Error message.")
		testLogger.success("Success message.")

	def testContextMethods(self):
		'''Displays all contextual methods.
		'''
		testLogger = Logger()

		#Verbose and debug depend on context.
		print("")
		testLogger.verbose("You shouldn't be able to see this!")
		testLogger.debug("You shouldn't be able to see this!")

		testLogger._minimumPrintableLevel = LogLevel.Verbose
		print("You should see a verbose message: ")
		testLogger.verbose("Verbose message.")
		print("You should see a debug message: ")
		testLogger.debug("Debug message.")

if __name__ == "__main__":
	suite = unittest.TestLoader().loadTestsFromTestCase(LogTest)
	unittest.TextTestRunner(verbosity=2).run(suite)



'''Dummy steps that return specified values.
'''

from ..setup import StepResult

class TestSteps(object):
	'''Valid dummy steps that return specified values.
	'''

	class Pass(object):
		'''Represents an operation
		that had to be performed and was done successfully.
		'''
		Name = "This step should pass installation"

		@classmethod
		def check(cls, _):
			'''Dummy check. Does nothing.
			'''
			return StepResult.NotDone

		@classmethod
		def run(cls, _):
			'''Dummy run. Does nothing.
			'''
			return StepResult.Success

	class Fail(object):
		'''Represents an operation
		that had to be performed but failed installation.
		'''
		Name = "This step should fail installation"

		@classmethod
		def check(cls, _):
			'''Dummy check. Does nothing.
			'''
			return StepResult.NotDone

		@classmethod
		def run(cls, _):
			'''Dummy run. Does nothing.
			'''
			return StepResult.Fail

	class Installed(object):
		'''Represents an operation
		that was already performed?
		TODO: Should we remove StepResult.Installed? How is this different from Success in practice?
		'''
		Name = "This step should be marked as installed"

		@classmethod
		def check(cls, _):
			'''Dummy check. Does nothing.
			'''
			return StepResult.Installed

		@classmethod
		def run(cls, _):
			'''Dummy run. Does nothing.
			'''
			return StepResult.Installed

	class Unnecessary(object):
		'''Represents an operation
		that was unnecessary and was skipped.
		'''
		Name = "This step should be skipped as unnecessary"

		@classmethod
		def check(cls, _):
			'''Dummy check. Does nothing.
			'''
			return StepResult.Unnecessary

		@classmethod
		def run(cls, _):
			'''Dummy run. Does nothing.
			'''
			return StepResult.Unnecessary

	class AlreadyDone(object):
		'''Represents an operation
		that was already performed and skipped.
		'''
		Name = "This step should be skipped as already being done"

		@classmethod
		def check(cls, _):
			'''Dummy check. Does nothing.
			'''
			return StepResult.AlreadyDone

		@classmethod
		def run(cls, _):
			'''Dummy run. Does nothing.
			'''
			return StepResult.AlreadyDone

	class NotDone(object):
		'''Represents an operation
		that had to be performed and was done successfully.
		TODO: Again, how is this different from TestSteps.Pass?
		'''
		Name = "This step should be checked as not done and run"

		@classmethod
		def check(cls, _):
			'''Dummy check. Does nothing.
			'''
			return StepResult.NotDone

		@classmethod
		def run(cls, _):
			'''Dummy run. Does nothing.
			'''
			print("In NotDone.run()")
			return StepResult.Success

	class WrongVersion(object):
		'''Represents an operation
		that had to be performed because the
		asset involved has the wrong version.
		'''
		Name = "This step should be checked as wrong version and run"

		@classmethod
		def check(cls, _):
			'''Dummy check. Does nothing.
			'''
			return StepResult.WrongVersion

		@classmethod
		def run(cls, _):
			'''Dummy run. Does nothing.
			'''
			print("In WrongVersion.run()")
			return StepResult.Success

	class InProgress(object):
		'''Represents an operation
		that is running but not complete.
		'''
		Name = "In-progress step; This step should be checked as not done and run"

		@classmethod
		def check(cls, _):
			'''Dummy check. Does nothing.
			'''
			return StepResult.InProgress

		@classmethod
		def run(cls, _):
			'''Dummy run. Does nothing.
			'''
			print("In InProgress.run()")
			return StepResult.Success

class AnomalousSteps(object):
	'''These steps all have something
	that should cause an error or warning.
	'''
	class Nameless(object):
		'''Step is missing its name,
		but is otherwise valid.
		'''

		@classmethod
		def check(cls, _):
			'''Dummy check. Does nothing.
			'''
			return StepResult.NotDone

		@classmethod
		def run(cls, _):
			'''Dummy run. Does nothing.
			'''
			print("In WrongVersion.run()")
			return StepResult.Success



'''Attempts to place the given notification to the user's
screen.
'''
import sys
import os

kMacOsNotifyCommand = "osascript -e 'display notification \"{1}\" with title \"{0}\"'"

def postNotification(title, body = ""):
	'''Sends an OS notification.

	:param str title: The notification's title.
	:param str body: The notification's body. Can be empty or None.
	'''
	if sys.platform == "linux" or sys.platform == "linux2":
		raise NotImplementedError("Notifications not implemented for Linux!")
	elif sys.platform == "darwin":
		#MacOS doesn't have a no-body notification.
		os.system(kMacOsNotifyCommand.format(title, body))
	elif sys.platform == "win32":
		raise NotImplementedError("Notifications not implemented for Windows!")



from .basis import Basis
from .setup import Flag
from .setup import Context, DistroCode, OSCode
from .setup import Logger
from .setup import Registry
from .setup import Version, VersionPair
from .setup import StepExecutor, StepResult, StepResultCategory, StepResultReason, ScopedStep
from .setup import PackageManager, ShellManager
from .setup import StepExecutorExceptionBase, StepExecutorError, StepExecutorWarning, Exceptions, CommandExecutionError
__all__ = ['basis', 'setup']



'''
Defines a log renderer that performs its
rendering by writing text to a given file path.
'''

class FilePoster(object):
	'''Renders log messages as OS notifications.
	'''

	def __init__(self, filePath):
		'''Class initializer.

		:param str filePath: The path to the file to save the log text in.

		:raises: Unknown, depends on the errors file.open() raises.
		'''
		raise NotImplementedError()

	def render(self, logText):
		'''Posts the log text as
		colored text to standard output.

		:param str logText: The text to be displayed.
		'''
		raise NotImplementedError()



'''Argument parsing system and classes.
'''

import argparse

class Flag(object):
	'''Represents a boolean flag.
	'''

	def __init__(self, description, value, shortValue=None):
		self.description = None
		self.name = None
		self.value = None
		self.shortValue = None
		if value is None or len(value) < 1:
			raise RuntimeError("Can't make a flag without a value")
		#Convert any dashes in the name to underscores
		#so it is a valid Python variable.
		self.name = value.replace("-", "_")
		if len(value) > 1:
			self.value = "--{0}".format(value)
		else:
			self.value = "-{0}".format(value)
		if shortValue is not None and len(shortValue) == 1 and shortValue != value:
			self.shortValue = "-{0}".format(shortValue)
		self.description = description

	def valueList(self):
		'''Returns all valid aliases for this flag.
		'''

		results = []
		if self.value is not None:
			results.append(self.value)
		if self.shortValue is not None:
			results.append(self.shortValue)
		return results

def _addBooleanFlags(parser, flags):
	for f in flags:
		parser.add_argument(*f.valueList(),
			action="store_true",
			help=f.description)

#Parsing constants.
class Flags(object):
	'''Predefined flags.
	'''
	_CheckOnlyDesc = "Makes installer only check for installed tools. Takes precedence over --verify."
	CheckOnly = Flag(_CheckOnlyDesc, "check-only", "c")
	Verbose = Flag("Enables verbose output.", "verbose", "v")
	Debug = Flag("Enables debug output.", "debug", "d")
	Verify = Flag("Runs a check after installation. Does nothing if --check-only is set.", "verify")
	PermitWarnings = Flag("If set, warning exceptions are not thrown", "permit-warnings", "p")
	AllBoolean = [CheckOnly, Verbose, Debug, Verify, PermitWarnings]

class Arguments(object):
	'''Parses command-line arguments.
	'''

	ProgramDescription = "Installs needed tools to build the product."
	ParserInitialized = False
	Parser = argparse.ArgumentParser(description=ProgramDescription)

	@classmethod
	def getParser(cls):
		'''Gets an instance of the parser.
		'''

		parser = cls.Parser
		if not cls.ParserInitialized:
			_addBooleanFlags(parser, Flags.AllBoolean)
			cls.ParserInitialized = True
		return parser

	@classmethod
	def parse(cls, args=None, extraFlags=()):
		'''Performs an argparse with the defined flags on
		the given argument list.
		Returns an argparse result object.
		'''

		parser = cls.getParser()
		#Add any extra arguments as needed.
		_addBooleanFlags(parser, extraFlags)
		result = None
		if args is None:
			result = parser.parse_args()
		else:
			result = parser.parse_args(args)
		return result



'''Converts a PBZX file to a TAR file.
Adapted from http://newosxbook.com/src.jl?tree=listings&file=pbzx.c
'''

import struct
import argparse

def bswap64(i):
	'''Does a binary swap on a uint_64.
	'''
	return struct.unpack("<Q", struct.pack(">Q", i))[0]

def asUint64(i):
	'''Converts read string to an unsigned 64-bit integer.
	'''
	return struct.unpack("Q", i)[0]

def pbzxToTar(inFile, outFile):
	'''Converts a PBZX file to a TAR file.
	Parameters:
		* inFile: The file object to unpack.
		* outFile: The destination file object to write output to.
	'''
	#Check it's a PBZX first.
	kPbzxMagic = '\x70\x62\x7A\x78' #pbzx
	fileMagic = inFile.read(4)
	if fileMagic != kPbzxMagic:
		raise RuntimeError("File doesn't have PBZX magic code; is this a PBZX file?")

	#If it is, start reading chunks.
	kUintSize = 8
	flags = 0
	length = 0

	#Get the first flag data.
	flags = bswap64(asUint64(inFile.read(kUintSize)))
	print("Flags: 0x%x" % flags)

	numBadChunks = 0
	numBadHeaders = 0
	numBadFooters = 0
	numBadLength = 0

	#While we have chunks...
	while flags & 0x01000000:
		flags = bswap64(asUint64(inFile.read(kUintSize)))
		#Limit chunk size to 4 GiB.
		kMaxChunkLength = 4 * (1 << 30)
		length = bswap64(asUint64(inFile.read(kUintSize)))

		if length > kMaxChunkLength:
			#Skip any chunks with bad length counts.
			numBadChunks += 1
			numBadLength += 1
		else:
			#Otherwise, go ahead and read.
			chunkBuffer = inFile.read(length)

			#We want the XZ header/footer if it's the payload,
			#but prepare_payload doesn't have that,
			#so just warn.
			header = chunkBuffer[:5]
			kExpectedHeader = "\xFD7zXZ"
			if header != kExpectedHeader:
				numBadChunks += 1
				numBadHeaders += 1
			else:
				#Check the footer too.
				#Footer is "YZ".
				footer = chunkBuffer[-2:]
				kExpectedFooter = "\x59\x5A"
				if footer != kExpectedFooter:
					numBadChunks += 1
					numBadFooters += 1
				else:
					#Only write out valid chunks.
					outFile.write(chunkBuffer)

	if numBadChunks > 0:
		#Report any unread chunks.
		errorString = "File had {0} unreadable chunks that weren't written:\n\tBad headers: {1}\n\tBad footers: {2}\n\tBad length counts: {3}".format(numBadChunks, numBadHeaders, numBadFooters, numBadLength)
		print(errorString)

		#But this is only an error if there
		#are bad footers or length counts; apparently
		#some non-data chunks can have invalid headers?
		if numBadFooters > 0 or numBadLength > 0:
			raise RuntimeError(errorString)

def main():
	'''Entry point when used as a program.
	'''
	parser = argparse.ArgumentParser(description="Converts a PBZX file to a TAR file.")
	parser.add_argument("inFile", help="The PBZX file to read.")
	parser.add_argument("outFile", help="The PBZX file to write.")
	args = parser.parse_args()

	#Try to open both files.
	print("Opening files:\n\tInput: {0}\n\tOutput: {1}".format(args.inFile, args.outFile))
	inFile = open(args.inFile, "rb")
	outFile = open(args.outFile, "wb")
	print("Files open, beginning conversion")

	pbzxToTar(inFile, outFile)
	print("Conversion complete.")

if __name__ == "__main__":
	main()



'''Unit tests for stepbase module.
'''

import unittest
from ...sampleteststeps import TestSteps, AnomalousSteps
from .shellmanager import ShellManagerTest
from .packagemanager import PackageManagerTest
from .scoped_step import ScopedStepTest
from ....setup import Context, Logger
from ....setup import StepExecutor, StepResult, StepResultReason, StepResultCategory
from ....setup import StepExecutorWarning

#These steps should return True for their install step.
SuccessSteps = ((TestSteps.Pass(), TestSteps.Installed(),
 TestSteps.NotDone(), TestSteps.InProgress(), TestSteps.WrongVersion()), StepResultCategory.Success)
#These steps should not run their install step.
SkipSteps = ((TestSteps.Unnecessary(), TestSteps.AlreadyDone()), StepResultCategory.Skip)
FailSteps = ((TestSteps.Fail(),), StepResultCategory.Fail)
AllSteps = (SuccessSteps, SkipSteps, FailSteps)

SuccessCheckSteps = ((TestSteps.Installed(),), StepResultCategory.Success)
#These steps should not run their install step.
SkipCheckSteps = ((TestSteps.Unnecessary(), TestSteps.AlreadyDone()), StepResultCategory.Skip)
FailCheckSteps = ((TestSteps.Pass(), TestSteps.Fail(), TestSteps.NotDone(), TestSteps.InProgress(), TestSteps.WrongVersion()), StepResultCategory.Fail)
AllCheckSteps = (SuccessCheckSteps, SkipCheckSteps, FailCheckSteps)

class StepExecutorTest(unittest.TestCase):

	def _validateVerb(self, stepExecutor, stepSetAndCategory, performMethodDescription, performMethod):
		'''Checks that the given method of StepExecutor
		performs under unit tests.
		Parmeters:
			* stepExecutor: The StepExecutor being tested.
			* check_only: Boolean. If True this runs unit tests
			as if context.check_only was set.
			* stepSetAndCategory: ((TestStep,), StepResultCategory). Lists the steps to be run and the expected category their results should fall under.
			If a step does not give a result in the expected category, that is a test failure.
			* performMethodDescription: A human-readable name for the method being run.
			* performMethod: The StepExecutor method to be tested.
		Returns:
			* True if performMethod passed all tests, False otherwise.
		Raises:
			* RuntimeError if performMethodDescription isn't set.
		'''
		assert isinstance(stepExecutor, StepExecutor)
		stepSet = stepSetAndCategory[0]
		expectedResultCategory = stepSetAndCategory[1]
		self.assertTrue(expectedResultCategory >= 0 and expectedResultCategory < StepResultCategory.Count,
			"Expected result category '{0}' not a valid StepResultCategory value".format(expectedResultCategory))
		print("")
		if not performMethodDescription:
			raise RuntimeError("_validateVerb(): performMethodDescription parameter can't be empty")
		#Test: Checkonly mode always returns True, even if the check failed.
		validated = True
		for step in stepSet:
			result = performMethod(step)
			if result.category != expectedResultCategory:
				print("Step {0} should return result in category {1} during a {2}, returned {3}".format(step.Name,
						StepResultCategory.Descriptions[expectedResultCategory],
						performMethodDescription,
						result))
				validated = False

		originalPermitWarnings = stepExecutor.context.permit_warnings
		#Also check that nameless steps throw...
		stepExecutor.context.permit_warnings = False
		try:
			performMethod(AnomalousSteps.Nameless)
			print("Method {0} was supposed to raise for nameless steps but didn't".format(performMethodDescription))
			validated = False
		except StepExecutorWarning:
			pass

		#But not when permit-warnings is set.
		stepExecutor.context.permit_warnings = True
		try:
			performMethod(AnomalousSteps.Nameless)
		except StepExecutorWarning:
			print("Method {0} wasn't supposed to raise for nameless steps when context.permit_warnings is True but did anyway".format(performMethodDescription))
			validated = False

		return validated

	def _validateCheck(self, check_only, stepSetAndCategory):
		logger = Logger()
		ctx = Context(logger)
		stepExecutor = StepExecutor(ctx, logger)
		ctx.check_only = check_only
		return self._validateVerb(stepExecutor, stepSetAndCategory, "check", stepExecutor.check)

	def _validatePerform(self, check_only, stepSetAndCategory):
		logger = Logger()
		ctx = Context(logger)
		stepExecutor = StepExecutor(ctx, logger)
		ctx.check_only = check_only
		return self._validateVerb(stepExecutor, stepSetAndCategory, "perform", stepExecutor.perform)

	def _validateAllSteps(self, check_only, allStepSetAndCategory, performMethodDescription, performMethod):
		allPassed = True
		for stepSetAndCategory in allStepSetAndCategory:
			category = stepSetAndCategory[1]
			categoryName = StepResultCategory.Descriptions[category]
			if not performMethod(check_only,
				stepSetAndCategory):
				allPassed = False
				print("Category {0} failed {1} validation when check_only = {2}".format(categoryName, performMethodDescription, check_only))
		if not allPassed:
			print("Not all steps passed {0} validation when check_only = {1}".format(performMethodDescription, check_only))
		return allPassed

	def _validateAllStepsAndStates(self, checkOnlyAndStepSets, performMethodDescription, performMethod):
		allPassed = True
		for state in checkOnlyAndStepSets:
			check_only = state[0]
			allStepSetAndCategory = state[1]
			allPassed = allPassed and self._validateAllSteps(check_only, allStepSetAndCategory, performMethodDescription, performMethod)
		return allPassed

	def testResultValues(self):
		'''Test that all Results have a description and value.
		'''
		self.assertFalse(StepResultReason.Count < len(StepResultReason.Descriptions),
			"A result reason is missing a description")
		self.assertFalse(StepResultReason.Count < len(StepResultReason.ToCategory),
			"A result reason is missing a category")
		self.assertFalse(StepResultReason.Count > len(StepResultReason.Descriptions),
			"There are too many result descriptions")
		self.assertFalse(StepResultReason.Count > len(StepResultReason.ToCategory),
			"There are too many mappings from a result to a category")

	def testFromBool(self):
		'''Test that StepResult.fromBool()
		returns expected values.
		'''

		self.assertTrue(StepResult.fromBool(True) == StepResult.Success)
		self.assertTrue(StepResult.fromBool(False) == StepResult.Fail)

	def testPerform(self):
		'''Test that StepExecutor.perform()
		returns expected values from steps and does not
		raise exceptions during execution.
		'''
		checkOnlyStates = ((True, AllCheckSteps), (False, AllSteps))
		self.assertTrue(self._validateAllStepsAndStates(checkOnlyStates, "perform", self._validatePerform))

	def testCheck(self):
		'''Test that StepExecutor.check()
		returns expected values from steps and does not
		raise exceptions during execution.
		'''
		checkOnlyStates = ((True, AllCheckSteps), (False, AllCheckSteps))
		self.assertTrue(self._validateAllStepsAndStates(checkOnlyStates, "check", self._validateCheck))


def suite():
	'''Returns all cases that should be run
	to test this part of the application.
	'''

	all_cases = (ShellManagerTest, PackageManagerTest, StepExecutorTest, ScopedStepTest)
	all_suites = []

	for case in all_cases:
		all_suites.append(unittest.TestLoader().loadTestsFromTestCase(case))

	return unittest.TestSuite(all_suites)

if __name__ == "__main__":
	unittest.TextTestRunner(verbosity=2).run(suite())



'''Contains unit tests for basis.setup.step.scoped_step.
'''
import unittest
from ..... import Basis, ScopedStep

class ScopedStepTest(unittest.TestCase):
	def __init__(self):
		unittest.TestCase.__init__(self)
		self.basis = Basis()

	def testSuccessfulStep(self):
		'''Test that a step that runs without any exceptions
		is considered successful in the Basis instance.
		'''
		numFailedStepsBefore = len(self.basis.failedSteps())

		with ScopedStep("Test step", self.basis) as _step:
			pass

		numFailedStepsAfter = len(self.basis.failedSteps())
		self.assertEquals(numFailedStepsBefore, numFailedStepsAfter)

	def testFailedStep(self):
		'''Test that a step that has an unhandled exception
		is considered failed in the Basis instance.
		'''
		numFailedStepsBefore = len(self.basis.failedSteps())

		try:
			with ScopedStep("Test step", self.basis) as _step:
				raise Exception("Test exception")
		except:
			pass

		numFailedStepsAfter = len(self.basis.failedSteps())
		self.assertGreater(numFailedStepsAfter, numFailedStepsBefore)


def suite():
	'''Returns all cases that should be run
	to test this part of the application.
	'''

	all_cases = (ScopedStepTest,)
	all_suites = []

	for case in all_cases:
		all_suites.append(unittest.TestLoader().loadTestsFromTestCase(case))

	return unittest.TestSuite(all_suites)

if __name__ == "__main__":
	unittest.TextTestRunner(verbosity=2).run(suite())



'''
Defines a log renderer that performs its
rendering by creating an OS notification.
'''
from .internal import postNotification

class NotificationPoster(object):
	'''Renders log messages as OS notifications.
	'''

	def __init__(self):
		raise NotImplementedError()

	def render(self, logText):
		'''Posts the log text as an OS notification.

		:param str logText: The text to be displayed.
		'''
		postNotification(logText)



'''
Defines a log renderer that performs its
rendering by writing colored text to standard output.
'''

class TerminalPoster(object):
	'''Renders log messages as OS notifications.
	'''

	def __init__(self):
		raise NotImplementedError()

	def render(self, logText):
		'''Posts the log text as
		colored text to standard output.

		:param str logText: The text to be displayed.
		'''
		raise NotImplementedError()



'''Defines the ScopedStep class.
'''
from .. import StepResult

kDefaultStepName = "Unnamed step"

class ScopedStep(object):
	'''Similar to Step, but is used as a context.
	Upon entering the ScopedStep's context, the step is marked as running,
	and upon exiting it is marked as finished.

	If an unhandled exception occurs,
	the step fails with its reason specified as the unhandled exception.
	'''
	def __init__(self, stepName, basis):
		'''Instance initializer.

		:param str stepName: The name of the step. If not provided, defaults to 'Unnamed step'.
		:param basis.Basis basis: The Basis instance to use this step on. 
		
		:raises RuntimeError: If basis is None.
		'''
		self.stepName = kDefaultStepName
		if stepName:
			self.stepName = stepName
		else:
			basis.logger.warning("Step has Name field but it was empty; was this intended?")

		if not basis:
			raise RuntimeError("ScopedStep needs context object to operate!")
		self.basis = basis
		self.step_status = StepResult.NotDone

	def __enter__(self):
		'''Context entry method.
		'''
		self.basis.logger.info("Performing step: " + self.stepName + "...")
		self.basis.step_status = StepResult.InProgress

	def __exit__(self, exc_type, exc_value, traceback):
		'''Context exit method.

		:param exc_type:	The exception type if an exception occured
							within the yield section, None otherwise.
		:param exc_value:	The exception object itself if an exception occured
							within the yield section, None otherwise.
		:param traceback:	The traceback to the function generating
							the unhandeled exception if an exception occured
							within the yield section, None otherwise.
		'''
		#Check: did this step have any exception?
		if not exc_value:
			#If not, mark this step as successful.
			self.step_status = StepResult.Success
			self.basis.logger.success("Step succeeded: " + self.stepName)
		else:
			#If an unhandled exception occurs,
			#mark this step as failed with
			#the unhandled exception specified as the reason.
			self.step_status = StepResult.Fail
			self.basis.logger.error("Step {0} failed install with an error: {1}".format(self.stepName, exc_value))
			self.basis.addFailedStep(self.stepName)


'''Defines step result values and the step executor.
'''
from .exceptions import StepExecutorError, Exceptions
from .packagemanager import PackageManager
from .shellmanager import ShellManager
from .stepresult import StepResult
from ..context import Context
from ..logging import Logger

class StepExecutor(object):
	'''Runs steps.
	'''

	def __init__(self, context, logger):
		if not context:
			raise RuntimeError("StepExecutor needs context object to operate!")
		if not context.isValid():
			raise RuntimeError("Trying to setup StepExecutor with invalid context!")
		self.context = context
		assert isinstance(self.context, Context)
		self.log = logger
		assert isinstance(self.log, Logger)
		self.shell = ShellManager(self.context, self.log)
		self.package = PackageManager(self.shell, self.log)

	def _stepNameFromStep(self, step):
		'''Gets the step name from a step.
		Returns:
			* The step name if step.Name exists and is not empty, "Unnamed step" otherwise.
		Raises:
			* StepExecutorWarning if step.Name does not exist.
		'''
		stepName = "Unnamed step"
		try:
			if step.Name:
				stepName = step.Name
			else:
				self.log.warning("Step has Name field but it was empty; was this intended?")
		except AttributeError:
			Exceptions.raiseAndLogStepWarning(self.log, self.context, "Step is missing Name field!")

		return stepName

	def check(self, step):
		'''
		Only checks if a step needs to be done.
		Parameters:
			* step: The step to check.
		Returns: A StepResult indicating if the step was
		successfully checked or not.
		Raises:
			* StepExecutorWarning if step is missing Name field and self.context.permit_warnings is False.
		'''
		stepName = self._stepNameFromStep(step)

		#Perform the actual check.
		try:
			step_status = step.check(self)
		except StepExecutorError as e:
			self.log.error("Step '{0}' failed check with an error: {1}".format(stepName, e))
			step_status = StepResult.Fail

		step_status_string = step_status.description
		if self.context.check_only:
			if step_status:
				self.log.success("{0}: {1}".format(step_status_string, stepName))
			else:
				self.log.warning("{0}: {1}".format(step_status_string, stepName))
				self.context.failedSteps.append(stepName)
		return step_status

	def perform(self, step):
		'''
		Checks if a step needs to be done, and if so
		performs the step.
		Parameters:
			- step: The step to run.
		Returns: A StepResult indicating if the step was
		successfully performed or not.
		'''
		step_status = self.check(step)
		step_status_string = step_status.description
		stepName = self._stepNameFromStep(step)

		#Perform the actual step.
		if not self.context.check_only:
			#See if we need to do this step.
			if step_status:
				self.log.success("{0}, skipping: {1}".format(step_status_string,
					stepName))
			else:
				#Perform the step.
				self.log.info("Performing step: " + stepName + "...")
				try:
					step_status = step.run(self)
					if step_status is None:
						#Like with check(), assume success if no result was returned.
						self.log.warning("Step {0} didn't \
						return a StepResult from its run() \
					method, assuming success".format(stepName))
						step_status = StepResult.Success
				except StepExecutorError as e:
					self.log.error("Step {0} failed install with an error: {1}".format(stepName, e))
					step_status = StepResult.Fail
				if not step_status:
					self.log.error("Step failed: " + stepName)
					self.context.failedSteps.append(stepName)
				else:
					self.log.success("Step succeeded: " + stepName)
		return step_status



'''Installs development tools for each platform.
'''
import traceback
from .setup.context import Context, OSCode
from .setup.step.stepbase import StepExecutor
from .setup.logging import Logger
from .setup.helpers import ShellHelpers

class BasisError(RuntimeError):
	'''Runtime error during Basis execution.
	'''
	pass

###
# ENTRY POINT
###
class Basis(object):
	'''Runs steps provided by implementor.
	'''

	def __init__(self, args=None, extraFlags=(), requireAdminFlags=(), customLogger=None):
		'''Initializer.
		Arguments:
			* `args`: The arguments to send to the argument parser as a single string.
				If None, this will take the initial invocation as the argument string.
			* `extraFlags`: A tuple of Flag objects that
				act as extra boolean flags used by the basisper.
				Default is the empty tuple (), throws BasisError if this is None.
			* `requireAdminFlags`: Specifies the admin mode as a tuple of OSCode flags.
				If no flags are set (empty tuple), no check is made;
				if a platform's flag is set and the client is running the given platform,
				the program will quit if the client's user is not an admin.
				Throws BasisError if this is None.
			* `customLogger`: A Logger instance to be used
				by the program for this run. If None,
				the Basis instance will make its own Logger.
		'''
		#Sanity check.
		if extraFlags is None:
			raise BasisError("Trying to make a Basis instance with extraFlags set to None!")
		if requireAdminFlags is None:
			raise BasisError("Trying to make a Basis instance with requireAdminFlags set to None!")

		#Arguments can be passed directly to the instance for testing.
		self.logger = Logger() if not customLogger else customLogger
		assert isinstance(self.logger, Logger)
		self._args = args
		self._extraFlags = extraFlags
		self._requireAdminFlags = requireAdminFlags
		#Message to print when installing.
		self.installStartMessage = "Beginning installation..."
		#Message to print when --check_only is set.
		self.checkStartMessage = "Beginning check..."
		#Message to print when --verify is set.
		self.verifyStartMessage = "Beginning post-install check..."
		self.installFailedMessage = "Installation failed on the following steps:"
		self.successMessage = "Installation complete."
		self._commonInit()

	def __del__(self):
		'''Deallocator.
		'''
		dumpPath = ""
		try:
			dumpPath = ShellHelpers.defaultLogFilePath(self._context)
			self.logger.dumpEntriesToFile(dumpPath)
		except Exception as e:
			self.logger.error("Failed to dump logs to '{0}': {1}! This run won't be saved to logs!".format(dumpPath, str(e)))

	def _commonInit(self):
		'''Sets up remaining fields for Basis.
		'''
		#Sanity check.
		if self._extraFlags is None:
			raise BasisError("Trying to run a Basis instance with extraFlags set to None!")
		if self._requireAdminFlags is None:
			raise BasisError("Trying to run a Basis instance with requireAdminFlags set to None!")

		#Check our platform.
		self._context = Context(self.logger, args=self._args, extraFlags=self._extraFlags)

		#If we're running verbose/in a check mode,
		#describe the context before running.
		if self._context.verbose or self._context.debug or self._context.check_only:
			self._context.printContextDescription()
		else:
			self._context.printContextSummary()
		print("")

		#The step executor instance.
		self.stepExecutor = StepExecutor(self._context, self.logger)
		#It's possible for the run to fail here if the script requires admin privileges;
		#ensure logging's ready first
		try:
			self._initLogging(self.stepExecutor, self.stepExecutor.context)
		except Exception as e:
			self.logger.error("Failed to initialize logging: {0}! This run might not be saved to logs!")

		#Make sure that we have appropriate privileges.
		osDescription = OSCode.Descriptions[self._context.osIdentifier]
		if self._context.osIdentifier in self._requireAdminFlags and (not self._context.userIsAdmin):
			raise BasisError("Script requires admin \
			privileges for current platform '{0}'\
			!".format(osDescription))
		#Also that this OS is supported!
		if not self._context.osSupported:
			raise BasisError("OS '{0}' not supported!".format(osDescription))

	def _initLogging(self, step, ctx):
		assert isinstance(step, StepExecutor)
		#If the log directory doesn't exist, add it
		loggingDir = ShellHelpers.loggingDirectory(ctx)
		try:
			if not step.shell.pathExists(loggingDir):
				self.logger.debug("Log directory '{0}' doesn't exist, creating it".format(loggingDir))
				step.shell.makeDir(loggingDir)
		except IOError as e:
			print("Couldn't setup logging directory '{0}' due to exception: {1}".format(loggingDir, str(e)))
			print("This run won't be saved to a log file!")

	def _preRun(self, step):
		#If we're installing...
		if not step.context.check_only:
			#Report that we're starting installation.
			if self.installStartMessage:
				self.logger.info(self.installStartMessage)
		#Otherwise, note we're just checking.
		else:
			if self.checkStartMessage:
				self.logger.info(self.checkStartMessage)

	def _cleanup(self, step):
		#Check if any steps failed...
		if step.context.failedSteps:
			#If this isn't check-only mode,
			#list the failed steps.
			if not step.context.check_only:
				self.logger.error(self.installFailedMessage)
				#If so, print any failed steps.
				for failedStep in step.context.failedSteps:
					self.logger.error("\t" + failedStep)
			#Exit with an error.
			return 1
		else:
			#Otherwise we're good!
			self.logger.success(self.successMessage)
			return 0

	def _verify(self, context, stepMethod, step):
		'''Secondary check done after the run is performed.
		Should not run if --check-only is set.
		'''
		if context.verify:
			#The context system should mark verify as False if check_only is True.
			#If both are True, someone has modified the context. Raise in this case.
			if context.check_only:
				errString = ("Verify mode and check-only mode are both somehow set;"
				" make sure that the context hasn't been manually modified.")
				raise BasisError(errString)
			self.logger.info(self.verifyStartMessage)
			originalCheckOnly = context.check_only
			context.check_only = True
			stepMethod(step)
			context.check_only = originalCheckOnly

	def _doRun(self, stepMethod, forceCheck):
		'''Main function for basis system.
		'''
		try:
			#If we recognize the OS, perform setup!
			#Do any pre-installation prep.
			self._preRun(self.stepExecutor)

			#Actually do the install/check.
			stepMethod(self.stepExecutor)

			#Clean up after the basis script.
			resultCode = self._cleanup(self.stepExecutor)

			#Run the post-install check if needed.
			self._verify(self._context, stepMethod, self.stepExecutor)

			return resultCode
		except BasisError:
			traceback.print_exc()
			return 1

	def run(self, stepMethod):
		'''Runs the provided step method,
		performing check() and run() on each step.
		'''
		return self._doRun(stepMethod, False)

	def check(self, stepMethod):
		'''Checks the provided step method;
		only check() is called on each step.
		'''
		return self._doRun(stepMethod, True)

	def addFailedStep(self, stepName):
		'''Adds a step to the list of steps
		that failed to run correctly on this Basis instance.

		:param str stepName: The name of the step that failed.
		'''
		self._context.failedSteps.append(stepName)

	def failedSteps(self):
		'''Returns a list of failed steps.

		:returns: A list of steps marked as failed in this Basis instance.
		'''
		return self._context.failedSteps[:]



'''Unit tests for shellmanager module.
'''
import os
import unittest
import platform
from ....setup import Context, Logger, LogLevel
from ....setup import ShellManager
from ....setup import StepExecutorError

TestArchiveName = "testzip.zip"
TestArchiveContentName = "testfile.txt"
TestDMGName = "example-dmg.dmg"
UnitTestPath = os.path.dirname(os.path.abspath(__file__))
TestArchivePath = UnitTestPath + "/" + TestArchiveName
TestArchiveContentPath = UnitTestPath + "/" + TestArchiveContentName
TestDMGPath = UnitTestPath + "/" + TestDMGName

class ShellManagerTest(unittest.TestCase):
	'''Unit tests for shellmanager module.
	'''

	@classmethod
	def setUpClass(cls):
		cls.log = Logger(LogLevel.Debug)
		cls.context = Context(cls.log)
		#Change the environment variable file
		#to a test file.
		cls.context.environmentVariableFile = "testEnvFile.txt"
		cls.shell = ShellManager(cls.context, cls.log)
		cls.DefaultShellCmd = ""
		#Turns out dir isn't an executable as far as distutil is concerned.
		cls.DefaultShellCmd = "whoami"
		cls.DefaultPath = ""
		cls.InvalidPath = ""
		if cls.context.osIsWindows():
			cls.DefaultPath = "C:\\Users"
			cls.InvalidPath = "Z:\\ritjhidjhrdiwe\\ergsgeswe"
		else:
			cls.DefaultPath = "/home"
			cls.InvalidPath = "/ritjhidjhrdiwe/ergsgeswe"
		cls.InvalidShellCmd = "rhsviudihdgius"

	def testInvokeShell(self):
		'''Test that this always returns an object
		and raises if None is passed.
		'''
		shellCmd = self.__class__.DefaultShellCmd
		invalidShellCmd = self.__class__.InvalidShellCmd
		shell = self.__class__.shell
		self.assertTrue(shellCmd)
		self.assertTrue(shell)

		self.assertTrue(shell._invokeShell(shellCmd) is not None)

		#Invalid calls should not raise.
		shell._invokeShell(invalidShellCmd)
		#Empty calls should raise.
		with self.assertRaises(StepExecutorError):
			shell._invokeShell("")
			self.log.error("ShellManager._invokeShell() didn't throw on empty input")
		with self.assertRaises(StepExecutorError):
			shell._invokeShell(None)
			self.log.error("ShellManager._invokeShell() didn't throw on None input")

	def testProgramExists(self):
		'''Test programExists().
		'''

		shellCmd = self.__class__.DefaultShellCmd
		shell = self.__class__.shell
		self.assertTrue(shellCmd)
		self.assertTrue(shell)

		#Ensure commands that exist return True...
		self.assertTrue(shell.programExists(shellCmd))
		#And commands that don't return False.
		self.assertFalse(shell.programExists(self.__class__.InvalidShellCmd))

	def testPathExists(self):
		'''Test pathExists().
		'''

		path = self.__class__.DefaultPath
		shell = self.__class__.shell
		self.assertTrue(shell)

		#Ensure paths that exist return True...
		self.assertTrue(shell.pathExists(path))
		#And paths that don't return False.
		self.assertFalse(shell.pathExists(self.__class__.InvalidPath))

	def testCommand(self):
		'''Test command().
		'''

		shellCmd = self.__class__.DefaultShellCmd
		invalidShellCmd = self.__class__.InvalidShellCmd
		shell = self.__class__.shell
		self.assertTrue(shellCmd)
		self.assertTrue(shell)

		#Should not raise.
		shell.command(shellCmd)

		#Invalid calls should raise.
		with self.assertRaises(StepExecutorError):
			shell.command(invalidShellCmd)
			self.log.error("ShellManager.command() didn't \
			throw on invalid input {0}".format(invalidShellCmd))
		#Empty calls should raise.
		with self.assertRaises(StepExecutorError):
			shell.command("")
			self.log.error("ShellManager.command() didn't throw on empty input")
		with self.assertRaises(StepExecutorError):
			shell.command(None)
			self.log.error("ShellManager.command() didn't throw on None input")

	def testMultiCommand(self):
		'''Test that multiCommand() terminates
		after running its list of commands
		and that invalid input raises an exception.
		'''
		shellCmd = self.__class__.DefaultShellCmd
		invalidShellCmd = self.__class__.InvalidShellCmd
		shell = self.__class__.shell
		self.assertTrue(shellCmd)
		self.assertTrue(shell)

		validCmdList = (shellCmd, shellCmd, shellCmd)
		invalidCmdList = (shellCmd, invalidShellCmd, shellCmd)
		emptyCmdList = ()

		#Should not throw.
		shell.multiCommand(validCmdList)
		#Invalid calls inside the multicommand shouldn't raise.
		shell.multiCommand(invalidCmdList)

		#Empty calls should raise.
		with self.assertRaises(StepExecutorError):
			shell.multiCommand(emptyCmdList)
			self.log.error("ShellManager.multiCommand() didn't throw on empty input")
		with self.assertRaises(StepExecutorError):
			shell.multiCommand(None)
			self.log.error("ShellManager.multiCommand() didn't throw on None input")
		#Non-tuple calls should raise.
		with self.assertRaises(StepExecutorError):
			shell.multiCommand("invalidcommand")
			self.log.error("ShellManager.multiCommand() didn't throw on non-tuple input")


	def testCommandWithOutput(self):
		'''Test commandWithOutput().
		'''

		shellCmd = self.__class__.DefaultShellCmd
		invalidShellCmd = self.__class__.InvalidShellCmd
		shell = self.__class__.shell
		self.assertTrue(shellCmd)
		self.assertTrue(shell)

		self.assertTrue(shell.commandWithOutput(shellCmd) is not None)
		#Empty calls should raise...
		with self.assertRaises(StepExecutorError):
			shell.commandWithOutput("")
			self.log.error("ShellManager.commandWithOutput() didn't throw on empty input")
		with self.assertRaises(StepExecutorError):
			shell.commandWithOutput(None)
			self.log.error("ShellManager.commandWithOutput() didn't throw on None input")
		#But invalid calls shouldn't.
		self.log.info("Next call should not raise exception...")
		shell.commandWithOutput(invalidShellCmd)

	def testParseShellCommand(self):
		'''Test parseShellCommand().
		'''

		context = self.__class__.context
		shell = self.__class__.shell

		self.assertTrue(context)
		self.assertTrue(shell)

		contextOriginalVerbose = context.verbose
		contextOriginalDebug = context.debug
		context.verbose = True
		context.debug = True
		#Test shell.parseCommand() by matching it against python version.
		actualVersion = platform.python_version()
		shellReportedVersion = shell.parseCommand("python --version", ".* (.*)").group(1)
		if actualVersion != shellReportedVersion:
			try:
				shellReportedVersion = shell.parseCommand("python3 --version", ".* (.*)").group(1)
			except RuntimeError:
				pass
		self.assertTrue(shellReportedVersion == actualVersion,
						"Parser should have found version '{0}', got '{1}' instead".format(actualVersion,
																							shellReportedVersion))
		#Also test that it raises on no match.
		badRegex = "vrivdiudiusbv"
		with self.assertRaises(StepExecutorError):
			shell.parseCommand("python --version", badRegex)
			self.log.error("ShellManager.parseCommand() didn't throw after no match")
		context.verbose = contextOriginalVerbose
		context.debug = contextOriginalDebug

	def testShowCommandOutput(self):
		'''Test that ShellManager.showCommandOutput
		properly displays output.
		'''

		shell = self.__class__.shell
		shellCmd = self.__class__.DefaultShellCmd
		self.assertTrue(shell)
		self.assertTrue(shellCmd)

		originalShowCommandOutput = shell.showCommandOutput
		shell.showCommandOutput = True
		self.log.info("ShellManager.showCommandOutput is \
		True; you should see output from the next commands.")
		shell.command(shellCmd)
		commandOutput = shell.commandWithOutput(shellCmd)
		self.assertTrue(commandOutput, "Didn't get any output from ShellManager.commandWithOutput()")
		commandOutput = shell.parseCommand(shellCmd, ".*")
		self.assertTrue(commandOutput, "Didn't get any output from ShellManager.parseCommand()")

		shell.showCommandOutput = False
		self.log.info("ShellManager.showCommandOutput is \
		False; you should NOT see printed output from the \
		next commands.")
		shell.command(shellCmd)
		commandOutput = shell.commandWithOutput(shellCmd)
		self.assertTrue(commandOutput, "Didn't get any output from ShellManager.commandWithOutput()")
		commandOutput = shell.parseCommand(shellCmd, ".*")
		self.assertTrue(commandOutput, "Didn't get any output from ShellManager.parseCommand()")

		#Restore shell state.
		shell.showCommandOutput = originalShowCommandOutput

	def testEnvironmentVariableCommands(self):
		'''Test the environment variable methods.
		'''

		shell = self.__class__.shell
		self.assertTrue(shell)

		#Change the environment variable file
		#to a test file.
		shell.context.environmentVariableFile = "testEnvFile.txt"
		#Make sure we're editing a nonexistent var (read it and make sure it's None)
		envVar = "EVIEOJMV"
		envVarValue = shell.getEnvironmentVariable(envVar)
		self.assertTrue(not envVarValue,
						"Environment variable {0} shouldn't exist but returned '{1}', aborting test".format(envVar,
						envVarValue))

		#Set it.
		envVarValue = "abcd"
		shell.setEnvironmentVariable(envVar, envVarValue)
		print("Environment variable {0} is now '{1}'".format(envVar,
			shell.getEnvironmentVariable(envVar)))

		#Append to it.
		envVarValue = "efgh"
		shell.appendEnvironmentVariable(envVar, envVarValue)
		print("Environment variable {0} is now '{1}'".format(envVar,
			shell.getEnvironmentVariable(envVar)))

		#Now clear it.
		envVarValue = ""
		shell.setEnvironmentVariable(envVar, envVarValue)
		print("Environment variable {0} is now '{1}'".format(envVar,
			shell.getEnvironmentVariable(envVar)))

		#Delete the file. It's okay if it doesn't exist.
		try:
			os.remove(shell.context.environmentVariableFile)
		except Exception:
			pass

	def testPrepareTempDirectory(self):
		'''Test prepareTempDirectory().
		'''

		shell = self.__class__.shell
		self.assertTrue(shell)

		shell.prepareTempDirectory()

	def testCleanupTempDirectory(self):
		'''Test cleanupTempDirectory().
		'''

		shell = self.__class__.shell
		self.assertTrue(shell)

		shell.cleanupTempDirectory()

	def testCopyAndRemove(self):
		'''Test copyFile() and removeFile().
		'''

		shell = self.__class__.shell
		self.assertTrue(shell)

		copyName = "copy.zip"
		copyPath = UnitTestPath + "/" + copyName
		shell.copyFile(TestArchivePath, copyPath)
		self.assertTrue(os.path.exists(copyPath))
		shell.removeFile(copyPath)
		self.assertTrue(not os.path.exists(copyPath))

	def testUnzip(self):
		'''Test unzip().
		'''

		shell = self.__class__.shell
		self.assertTrue(shell)

		shell.unzip(TestArchivePath, UnitTestPath)
		self.assertTrue(os.path.exists(TestArchiveContentPath))
		shell.removeFile(TestArchiveContentPath)

	def testDownloadFile(self):
		'''Test downloadFile().
		'''
		shell = self.__class__.shell
		assert isinstance(shell, ShellManager)

		sourcePath = "http://httpbin.org/html"
		destPath = os.path.basename(sourcePath)
		shell.downloadFile(sourcePath)
		#Assert that the file actually was downloaded...
		self.assertTrue(shell.pathExists(destPath))
		#Also that it's nontrivial in size.
		self.assertTrue(os.path.getsize(destPath) > 1)
		shell.removeFile(destPath)

	def testDMG(self):
		'''Test openDMG() and closeDMG().
		'''
		shell = self.__class__.shell
		assert isinstance(shell, ShellManager)
		#Skip if this isn't running under macOS.
		if not shell.context.osIsMacOS():
			self.skipTest("Can't test openDMG()/closeDMG() when not under macOS")

		#Otherwise...
		#See that we can open and close the DMG.
		mountPath = shell.openDMG(TestDMGPath)
		shell.closeDMG(mountPath)

if __name__ == "__main__":
	suite = unittest.TestLoader().loadTestsFromTestCase(ShellManagerTest)
	unittest.TextTestRunner(verbosity=2).run(suite)



'''Abstracts common shell commands.
'''

import os
import sys
import subprocess
import shutil
import re
import distutils.spawn
import zipfile
from .exceptions import StepExecutorError, ShellError, CommandExecutionError
from .. import Context
from .. import pbzx
from ..helpers import ShellHelpers

try:
	basestring
except NameError:
	basestring = str

TempDirectoryName = "temp"
TempDirectoryPath = ShellHelpers.absolutePath(TempDirectoryName)

#Disable R0201 since this is supposed to be an abstraction layer.
#pylint: disable=R0201
class ShellManager(object):
	'''Abstracts common shell commands.
	'''
	def __init__(self, ctx, logger):
		if not ctx:
			raise RuntimeError("ShellManager needs context object to operate")
		if not ctx.isValid():
			raise RuntimeError("Trying to setup ShellManager with invalid context!")
		self.context = ctx
		#If True, the standard output of any command*() call
		#is printed.
		self.showCommandOutput = True
		#If True, shell calls are escalated by default.
		self.escalateShellCalls = False
		#Null file object, for when output shouldn't be printed.
		self._nullFile = open(os.devnull, 'w')
		self._log = logger

	def _getStdoutFile(self):
		'''Returns the file descriptor
		that should be used for standard output/
		standard error.
		'''
		if self.showCommandOutput:
			return None
		return self._nullFile

	def _invokeShell(self, command, escalate=None, hideOutput=False, interactive=False, secure=False, shell=True):
		'''Base shell invocation method used by other methods.
		Returns the subprocess as a Popen object.
		Raises ShellError if command is empty or None.
		Options:
			* escalate: If True, runs command as administrator.
			* hideOutput: If True, output isn't echoed to screen.
			* interactive: If True, user can send input via standard input.
			* secure: If True, command invocation will not be echoed in any logging,
			not just the screen standard output.
			* shell: If True, is run within the current shell,
			allowing features such as piping but also making the call less secure.
		'''
		if not command:
			raise ShellError("No command given")
		if escalate is None:
			escalate = self.escalateShellCalls
		if escalate:
			command = self.context.escalateTool.format(command)
		if not secure:
			self._log.verbose("Shell command: " + command, self.context)
		# Add this when under Windows so a new terminal doesn't pop up.
		#startupinfo = subprocess.STARTUPINFO()
		#startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
		outFile = self._getStdoutFile()
		if hideOutput:
			outFile = subprocess.PIPE
		inFile = subprocess.PIPE
		if interactive:
			inFile = None
		exe = None
		if self.context.osIsLinux() or self.context.osIsMacOS():
			exe = "/bin/bash"
		elif self.context.osIsWindows():
			powershellPath = "C:\\Windows\\System32\\WindowsPowerShell\\v1.0\\powershell.exe"
			#Surround any ampersands with quotes.
			command = command.replace("&", "\"&\"")
			#...But don't replace double ampersands,
			#since those are almost certainly for a bash call.
			command = command.replace("\"&\"\"&\"", "&&")
			command = "{0} {1}".format(powershellPath, command)
		return subprocess.Popen(command,
							stdout=outFile,
							stderr=outFile,
							stdin=inFile,
							shell=True,
							executable=exe)

	def programExists(self, program):
		'''Returns True if the given program can be
		typed as a command and recognized by the shell
		without referring to its full path,
		False otherwise.
		Parameters:
			* program: The program name as a string.
		'''
		if self.context.osIsLinux() or self.context.osIsMacOS():
			try:
				self.command("which " + program, hideOutput=True)
				return True
			except StepExecutorError:
				return False
		elif self.context.osIsWindows():
			try:
				self.command("get-command {0} -CommandType Application".format(program), hideOutput=True)
				return True
			except StepExecutorError:
				return False
		return distutils.spawn.find_executable(program) is not None

	def pathExists(self, path):
		'''Returns True if the given path exists,
		False otherwise.
		Parameters:
			* path: The path name as a string.
		'''
		path = self.normalizedPath(path)
		result = os.path.exists(path)
		logString = "Path exists: {0}"
		if not result:
			logString = "Path doesn't exist: {0}"
		self._log.verbose(logString.format(path), self.context)
		return result

	def normalizedPath(self, path):
		'''Returns the path with all variables expanded and converted to the system's
		normalized path system.
		'''
		return ShellHelpers.absolutePath(path)

	def command(self, command, escalate=None, hideOutput=False, secure=False, shell=True):
		'''Performs the given command,
		raising a CommandExecutionError if the command
		doesn't return an exit code of 0.
		Options:
			* escalate: If True, runs command as administrator
			(via sudo on POSIX and runas on Windows).
			* hideOutput: If True, output isn't echoed to screen.
			* secure: If True, command invocation will not be echoed in any logging.
		'''
		commandResults = self._invokeShell(command, escalate, hideOutput, interactive=True, secure=secure, shell=shell)
		commandResults.communicate()
		returnCode = commandResults.returncode
		if returnCode != 0:
			errString = "Command '{0}' failed with return code {1}"
			raise CommandExecutionError(errString.format(command, returnCode), returnCode)

	def multiCommand(self, commandList, shell=True):
		'''Runs a list of commands in the specified order.
		Raises:
			* ShellError if:
				* commandList is empty or None.
				* commandList is not an iterable list/tuple.
		'''
		#Sanity check.
		if not commandList:
			raise ShellError("ShellManager.multiCommand(): \
			No commands given")
		if not hasattr(commandList, "__iter__"):
			raise ShellError("ShellManager.multiCommand(): \
			Command list '{0}' isn't iterable".format(commandList))
		if isinstance(commandList, basestring):
			raise ShellError("ShellManager.multiCommand(): \
			Command list '{0}' is a string; did you mean to \
			call command() instead?".format(commandList))

		self._log.info("Running multiple commands!")
		#Spawn a separate process and start the command execution.
		process = self._invokeShell(self.context.newShellTool)
		for command in commandList:
			#Run the command.
			#Remember the newline so it actually executes.
			#Python 3: This needs to explicitly be bytes!
			finalCommand = command
			if not shell:
				finalCommand = " ".join(finalCommand)
			process.stdin.write("{0}\n".format(finalCommand).encode('utf-8'))
		#Now close out the process. Return code isn't really useful
		#since it returns the value of the process, which is the shell we opened
		#and not the commands *in* the shell.
		process.communicate()

	def commandWithOutput(self, command, escalate=None, shell=True):
		'''Performs the given command and returns its
		standard output instead of its status.
		If escalate=True, runs command as administrator
		(via sudo on POSIX and runas on Windows).
		'''
		allchannels = self._invokeShell(command, escalate, hideOutput=True, shell=shell).communicate()
		result = allchannels[0]
		#If this is empty, use standard error.
		if not result:
			result = allchannels[1]

		#If this is Python 3, decode into a normal string.
		if sys.version_info[0] >= 3:
			result = result.decode()

		self._log.debug("Output before strip: {0}".format(result), self.context)
		self._log.debug("Standard error: {0}".format(allchannels[1]), self.context)
		if result:
			result = result.strip()
		return result

	def parseCommand(self, command, regex, escalate=None, shell=True):
		'''Performs the given command and runs a re.search()
		on its standard output.
		Returns the match results of the re.search() call.
		Raises a ShellError if no matches were found.
		'''
		infoString = "Matching on command \"{0}\" with regex \"{1}\""
		self._log.verbose(infoString.format(command, regex), self.context)
		shellOutput = self.commandWithOutput(command, escalate, shell)
		self._log.debug("Shell output: {0}".format(shellOutput), self.context)
		if not shellOutput:
			errString = "No output from command {0}, can't parse"
			raise ShellError(errString.format(shellOutput))
		#Do the actual search.
		result = re.search(regex, shellOutput)

		#If we're in debug mode, report if we matched.
		matchCountString = "No match"
		matchResultString = ""
		if result:
			matchCountString = "Match with {0} groups".format(len(result.groups()))
			#Print the entire matched string.
			matchResultString = ": {0}".format(result.group(0))
		debugMessage = "{0} found for regex {1}{2}".format(matchCountString, regex, matchResultString)
		self._log.debug(debugMessage, self.context)
		if result is None:
			raise ShellError(debugMessage)

		return result

	def secureCommand(self, command, escalate=None, shell=True):
		'''Runs a command without printing what the command was
		to output.
		'''
		self.command(command, escalate=escalate, secure=True, shell=shell)

	#Actual commands start here.

	def deletePath(self, path):
		'''Deletes the given path,
		if it exists.
		'''
		path = self.normalizedPath(path)
		if os.path.exists(path):
			try:
				self._log.warning("Removing path " + path)
				shutil.rmtree(path)
			except Exception as e:
				errString = "Couldn't remove path \"{0}\": {1}"
				raise ShellError(errString.format(path, e))

	def cleanupTempDirectory(self):
		'''Deletes the temporary storage directory,
		if it exists.
		'''
		self.deletePath(TempDirectoryPath)

	def prepareTempDirectory(self):
		'''Creates a temporary storage directory.
		'''
		self.cleanupTempDirectory()
		self.makeDir(TempDirectoryPath)

	def makeDir(self, dirPath):
		'''Creates the given directory and any
		directories above it that do not already exist.
		Throws a ShellError if the command fails
		for any reason.
		'''
		dirPath = self.normalizedPath(dirPath)
		try:
			self._log.verbose("Creating directory " + dirPath)
			os.makedirs(dirPath, exist_ok=True)
		except Exception as e:
			errString = "Couldn't create directory \"{0}\": {1}"
			raise ShellError(errString.format(dirPath, e))

	def getEnvironmentVariable(self, name):
		'''Returns the value of the given
		environment variable.
		'''

		#Under Windows:
		if self.context.osIsWindows():
			result = self.commandWithOutput("[Environment]::GetEnvironmentVariable(\\\"{0}\\\",\\\"User\\\")".format(name))
			self._log.verbose("VarName = '$env:{0}', Result = '{1}'".format(name, result), self.context)
			if not result:
				return ""
			return result
		#Under anything else:
		else:
			#Probably an echo command?
			return self.commandWithOutput("printenv {0}".format(name))

	def appendEnvironmentVariable(self, name, value):
		'''Sets the value of the given
		environment variable; if a value already exists
		for the variable, appends the given value
		to the variable.
		Returns True if the value could be set,
		False otherwise.
		'''
		initialValue = self.getEnvironmentVariable(name)
		finalValue = value
		if initialValue:
			#Under Windows:
			if self.context.osIsWindows():
				finalValue = "{0};{1}".format(initialValue, value)
			else:
				#Under anything else:
				finalValue = "{0}:{1}".format(initialValue, value)
		self._log.verbose("appendEnvironmentVariable(): Got initial value '{0}', final value is '{1}'".format(initialValue, finalValue), self.context)
		return self.setEnvironmentVariable(name, finalValue)

	def setEnvironmentVariable(self, name, value):
		'''Sets the value of the given
		environment variable.
		Raises a ShellError under Linux and macOS
		if there is no environment variable file set.
		'''
		if not name:
			raise RuntimeError("setEnvironmentVariable(): no environment value given")
		if value is None:
			raise RuntimeError("setEnvironmentVariable(): no value for environment value given")

		self._log.warning("Setting environment variable {0} to '{1}'".format(name, value))
		#Under Windows:
		if self.context.osIsWindows():
			self.command("[Environment]::SetEnvironmentVariable(\\\"{0}\\\", \\\"{1}\\\", \\\"User\\\")".format(name, value))
		#Under anything else:
		else:
			#Append to the user's .bashrc.
			bashCmd = "export {0}=\"{1}\"".format(name, value)
			envVarFile = self.context.environmentVariableFile
			if not envVarFile:
				errString = "Environment variable file '{0}' doesn't exist, can't modify environment variables"
				raise ShellError(errString.format(envVarFile))
			self.command("echo '{0}' >> {1}".format(bashCmd, envVarFile))
			#Also export now so this shell sees it.
			self.command("source {0}".format(envVarFile))

	def copyFile(self, srcPath, destPath):
		'''Copies a file from the source path to the destination path,
		overwriting anything in the destination.
		Raises a ShellError if srcPath and destPath are the same.
		'''
		srcPath = self.normalizedPath(srcPath)
		destPath = self.normalizedPath(destPath)
		if srcPath == destPath:
			raise ShellError("Trying to move '{0}' to itself".format(srcPath))
		shutil.copyfile(srcPath, destPath)

	def removeFile(self, path):
		'''Deletes the specified file.
		Doesn't work on directories.
		'''
		path = self.normalizedPath(path)
		os.remove(path)

	def unzip(self, zipPath, destPath):
		'''Unzips a ZIP archive to the destination path.
		'''
		zipPath = self.normalizedPath(zipPath)
		destPath = self.normalizedPath(destPath)
		archive = zipfile.ZipFile(zipPath, "r")
		self._log.verbose("Unzipping '{0}' to '{1}'".format(zipPath, destPath), self.context)
		archive.extractall(destPath)
		archive.close()

	def unpackXip(self, path):
		'''Unpacks a XIP archive in place.
		Only runs on macOS.
		Raises ShellError if run on a non-macOS system
		or if the file could not be unpacked for any reason.
		TODO: Add test coverage!
		'''
		if not self.context.osIsMacOS():
			raise ShellError("Can't unpack XIP files on a non-macOS system!")

		try:
			normPath = self.normalizedPath(path)
			baseFolder = os.path.dirname(normPath)
			fileName = os.path.basename(normPath)

			#Make a temporary folder to hold the file contents.
			tempFolder = baseFolder + "/basis-temp-c40595m0te"
			self._log.verbose("Creating temporary folders for XIP output.", self.context)
			self.command("mkdir {0}".format(tempFolder))
			#Move the archive into the temp folder.
			self.command("mv {0} {1}".format(normPath, tempFolder))
			tempInPath = tempFolder + "/" + fileName

			#Get the packed content data first.
			try:
				self._log.verbose("Retrieving packed XIP content...", self.context)
				"pkgutil --check-signature \"{0}\" && xar -xf \"{0}\"".format(tempInPath)
			except Exception as e:
				#Move the file back out first!
				self.command("mv {0} {1}".format(tempInPath, normPath))
				raise e
			#Now move the file back out.
			self.command("mv {0} {1}".format(tempInPath, normPath))

			#The data we want is in "./Content".
			inFile = open(baseFolder + "/Content", "rb")
			tempTarName = "/basis.unpackXip.tmp0v4g.tar"
			tempOutPath = tempFolder + tempTarName
			outFile = open(tempOutPath, "wb")
			self._log.verbose("Converting XIP to TAR...", self.context)
			#Convert to TAR.
			pbzx.pbzxToTar(inFile, outFile)
			#Now move the resultant TAR...
			self.command("mv {0} {1}".format(tempOutPath, baseFolder))
			finalOutPath = baseFolder + tempTarName
			self._log.verbose("Unpacking TAR...", self.context)
			#And unpack it.
			self.command("sudo tar x --strip-components=1 < {0}".format(finalOutPath))
			#Remove all temp files.
			self._log.verbose("XIP unpack complete, removing temporary content.", self.context)
			self.command("rm -rf {0}".format(finalOutPath))
			self.command("rm -rf {0}".format(tempFolder))

		except RuntimeError as e:
			#Rewrap any errors from the unpack function.
			raise ShellError(e.message)

	def downloadFile(self, remotePath, destPath=None, overwrite=False, escalate=None):
		'''Attempts to download a file from the given URL. If `destpath` is not provided,
		`remotePath`'s base will be used as the destination filename in the current working directory.
		If `overwrite` is True, any existing file at `destPath` will be removed; if a file does exist
		but `overwrite` is False, a ShellError will be raised instead.

		Raises ShellError if remotePath is not a valid string,
		if destPath could not be created for any reason,
		or if destPath already exists and `overwrite` is not True.
		'''
		assert isinstance(self.context, Context)

		#Sanity check.
		if not remotePath:
			raise ShellError("downloadFile() needs a URL to download from!")

		#Generate destPath if it wasn't provided.
		if not destPath:
			destPath = os.path.basename(remotePath)
			#This could be empty (if, say, you provide "http://google.com/").
			#In that case, provide a dummy name.
			if not destPath:
				destPath = "downloaded_file"

		self._log.verbose("Downloading file '{0}' to '{1}'".format(remotePath, destPath), self.context)

		#If the destination file already exists, overwrite or fail.
		if self.pathExists(destPath):
			if overwrite:
				self._log.verbose("File '{0}' already exists, \
				removing for download".format(destPath), 
				self.context)
				self.removeFile(destPath)
			else:
				raise ShellError("File '{0}' already exists \
				and overwrite flag not set, can't \
				download!".format(destPath))

		#Actually download the file...
		self.command(self.context.downloadTool.format(remotePath, destPath), escalate=escalate)

	def openDMG(self, dmgPath):
		'''Mounts the given DMG.
		Returns:
			* The path to the DMG's mount point as a string.
		Raises:
			* A StepExecutorError if called under a non-macOS
		platform or if the path could not be mounted.
		'''
		if not self.context.osIsMacOS():
			raise ShellError("openDMG() is only supported under macOS!")

		#Open the DMG.
		regex = "(\\S+)\\s+\\S+\\s+(\\S+)$"
		mountCmd = "hdiutil attach {0}".format(dmgPath)
		#Keep the disk's device and mount point.
		matches = self.parseCommand(mountCmd, regex)
		#dmgDevice = matches.group(1)
		mountPoint = matches.group(2)
		return mountPoint

	def closeDMG(self, mountPath):
		'''Ejects a mounted DMG at the given path.
		Raises:
			* A StepExecutorError if called under a non-macOS
		platform or if the path could not be ejected.
		'''
		if not self.context.osIsMacOS():
			raise ShellError("closeDMG() is only supported under macOS!")

		self.command("hdiutil detach " + mountPath)
